from typing import Text
import hashlib
import urllib.parse
from sqlalchemy.orm import Session
import pytest
from rasax.community import metrics
from rasax.community import config
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.story_service import StoryService
from rasax.community.constants import CONFIG_METRICS_ENABLED
from rasax.community.version import __version__


def test_get_start_event(
    session: Session,
    domain_service: DomainService,
    data_service: DataService,
    story_service: StoryService,
):
    domain = domain_service.get_or_create_domain()
    nlu_data = data_service.get_nlu_training_data_object()
    stories = story_service.fetch_stories()

    actual = metrics._get_rasa_x_start_event(session, "test", domain, nlu_data, stories)

    assert actual == {
        "project": "test",
        "rasa_x": __version__,
        "num_intent_examples": 0,
        "num_entity_examples": 0,
        "num_actions": 0,
        "num_templates": 0,
        "num_slots": 0,
        "num_forms": 0,
        "num_intents": 0,
        "num_entities": 0,
        "num_stories": 0,
        "num_conversations": 0,
        "num_models": 0,
        "num_lookup_table_files": 0,
        "num_lookup_table_entries": 0,
        "num_synonyms": 0,
        "num_regexes": 0,
    }


@pytest.mark.parametrize(
    "uuid, enabled", [("1234", True), ("5432", False), ("12344", True)]
)
def test_set_metrics_config_from_env_with_uuid(uuid: Text, enabled: bool):
    config.metrics_consent_given = enabled
    config.metrics_user_id = uuid
    metrics.set_metrics_config_from_env()

    assert metrics.are_metrics_enabled() == enabled
    assert metrics.metrics_id() == uuid


def test_set_metrics_config_from_env_without_uuid():
    config.metrics_user_id = None
    metrics.set_metrics_config_from_env()

    assert len(metrics.metrics_id()) >= 32


@pytest.fixture
def enable_metrics():
    previous_state = config.metrics_collection_config.get(CONFIG_METRICS_ENABLED, False)
    config.metrics_collection_config[CONFIG_METRICS_ENABLED] = True
    yield
    config.metrics_collection_config[CONFIG_METRICS_ENABLED] = previous_state


@pytest.fixture
def metrics_events(enable_metrics, monkeypatch):
    events = []

    def fn(_, event_name, properties=None, context=None):
        if properties:
            properties.pop("metrics_id", None)

        events.append(
            {"event": event_name, "properties": properties, "context": context}
        )

    monkeypatch.setattr(metrics, "_safe_track", fn)
    return events


@pytest.mark.parametrize(
    "referrer", ["http://localhost/talk", "http://localhost/talk/strict"]
)
def test_track_story_created(metrics_events, referrer):
    metrics.track_story_created(referrer)

    assert metrics_events[0] == {
        "event": "Story Created",
        "properties": {"story_created_from": urllib.parse.urlparse(referrer).path},
        "context": None,
    }


@pytest.mark.parametrize(
    "username", ["me", "test-user", "3d1693ea-9c51-4410-9e31-444e8a35727d"]
)
def test_track_message_received(metrics_events, username):
    metrics.track_message_received(username, "rasa")

    assert metrics_events[0] == {
        "event": "Message Received",
        "properties": {
            "username": hashlib.sha256(username.encode("utf-8")).hexdigest(),
            "channel": "rasa",
        },
        "context": None,
    }
